// This file is part of cxsd, copyright (c) 2015-2016 BusFaster Ltd.
// Released under the MIT license, see LICENSE.
"use strict";
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
var Exporter_1 = require('./Exporter');
var Namespace_1 = require('../Namespace');
/** Export parsed schema to a TypeScript d.ts definition file. */
var TS = (function (_super) {
    __extends(TS, _super);
    function TS() {
        _super.apply(this, arguments);
        this.construct = TS;
    }
    /** Format an XSD annotation as JSDoc. */
    TS.formatComment = function (indent, comment) {
        var lineList = comment.split('\n');
        var lineCount = lineList.length;
        var blankCount = 0;
        var contentCount = 0;
        var output = [];
        var prefix = '/\**';
        for (var _i = 0, lineList_1 = lineList; _i < lineList_1.length; _i++) {
            var line = lineList_1[_i];
            // Remove leading whitespace.
            line = line.replace(/^\s+/, '');
            // Remove trailing whitespace.
            line = line.replace(/\s+$/, '');
            if (!line)
                ++blankCount;
            else {
                if (blankCount && contentCount)
                    output.push(indent + prefix);
                output.push(indent + prefix + ' ' + line);
                prefix = '  *';
                ++contentCount;
                blankCount = 0;
            }
        }
        if (output.length)
            output[output.length - 1] += ' *\/';
        return (output.join('\n'));
    };
    TS.prototype.writeImport = function (shortName, relativePath, absolutePath) {
        return ('import * as ' +
            shortName +
            ' from ' +
            "'" + relativePath + "'" +
            ';');
    };
    /** Output list of original schema file locations. */
    TS.prototype.exportSourceList = function (sourceList) {
        var output = [];
        output.push('// Source files:');
        for (var _i = 0, sourceList_1 = sourceList; _i < sourceList_1.length; _i++) {
            var urlRemote = sourceList_1[_i];
            output.push('// ' + urlRemote);
        }
        output.push('');
        return (output);
    };
    TS.prototype.writeTypeRef = function (type, namePrefix) {
        var output = [];
        var namespace = type.namespace;
        var name = namePrefix + type.safeName;
        if (!namespace || namespace == this.namespace) {
            output.push(name);
        }
        else {
            // Type from another, imported namespace.
            var short = this.namespace.getShortRef(namespace.id);
            if (short) {
                output.push(short + '.' + name);
            }
            else {
                console.error('MISSING IMPORT ' + namespace.name + ' for type ' + type.name);
                output.push('any');
            }
        }
        return (output.join(''));
    };
    TS.prototype.writeMember = function (member, isExported) {
        var _this = this;
        var output = [];
        var comment = member.comment;
        var indent = isExported ? '' : '\t';
        var exportPrefix = isExported ? 'export var ' : '';
        if (comment) {
            output.push(TS.formatComment(indent, comment));
            output.push('\n');
        }
        output.push(indent + exportPrefix + member.safeName);
        if (!isExported && member.min == 0)
            output.push('?');
        output.push(': ');
        var outTypeList = member.typeList.map(function (type) {
            if (type.literalType && (!type.primitiveList || !type.primitiveList.length)) {
                return (type.literalType.name);
            }
            else
                return (_this.writeTypeRef(type, ''));
        });
        if (outTypeList.length == 0)
            return ('');
        var outTypes = outTypeList.sort().join(' | ');
        var suffix = '';
        if (member.max > 1)
            suffix = '[]';
        if (suffix && outTypeList.length > 1)
            outTypes = '(' + outTypes + ')';
        output.push(outTypes);
        output.push(suffix + ';');
        return (output.join(''));
    };
    TS.prototype.writeTypeContent = function (type) {
        var output = [];
        if (type.literalType) {
            var primitiveList = type.primitiveList;
            if (primitiveList && primitiveList.length > 0) {
                if (primitiveList.length > 1) {
                    output.push('(' + primitiveList.join(' | ') + ')');
                }
                else
                    output.push(primitiveList[0]);
            }
            else
                output.push(type.literalType.name);
        }
        else {
            var outMemberList = [];
            var output = [];
            var parentType = type.parent;
            for (var _i = 0, _a = type.attributeList; _i < _a.length; _i++) {
                var attribute = _a[_i];
                var outAttribute = this.writeMember(attribute, false);
                if (outAttribute)
                    outMemberList.push(outAttribute);
            }
            for (var _b = 0, _c = type.childList; _b < _c.length; _b++) {
                var child = _c[_b];
                var outChild = this.writeMember(child, false);
                if (outChild)
                    outMemberList.push(outChild);
            }
            output.push('{');
            if (outMemberList.length) {
                output.push('\n');
                output.push(outMemberList.join('\n'));
                output.push('\n');
            }
            output.push('}');
        }
        return (output.join(''));
    };
    TS.prototype.writeType = function (type, visible) {
        var namespace = this.namespace;
        var output = [];
        var comment = type.comment;
        var parentDef = '';
        var exportPrefix = visible ? 'export ' : '';
        var name = type.safeName;
        if (comment) {
            output.push(TS.formatComment('', comment));
            output.push('\n');
        }
        var content = this.writeTypeContent(type);
        if (namespace.isPrimitiveSpace) {
            output.push(exportPrefix + 'interface _' + name + ' { ' + 'content' + ': ' + type.literalType.name + '; }' + '\n');
        }
        else if (type.literalType) {
            parentDef = this.writeTypeRef(type.parent, '_');
            output.push(exportPrefix + 'type ' + name + ' = ' + content + ';' + '\n');
            if (type.primitiveList && type.primitiveList.length) {
                output.push('interface _' + name + ' extends ' + parentDef + ' { ' + 'content' + ': ' + name + '; }' + '\n');
            }
            else {
                output.push('type _' + name + ' = ' + parentDef + ';' + '\n');
            }
        }
        else {
            if (type.parent) {
                parentDef = ' extends ' + this.writeTypeRef(type.parent, '_');
            }
            output.push('interface _' + name + parentDef + ' ' + content + '\n');
            //output.push(exportPrefix + 'interface ' + name + ' extends _' + name + ' { new(): ' + name + '; }' + '\n');
            output.push(exportPrefix + 'interface ' + name + ' extends _' + name + ' { constructor: { new(): ' + name + ' }; }' + '\n');
            if (visible)
                output.push(exportPrefix + 'var ' + name + ': { new(): ' + name + ' };' + '\n');
        }
        return (output.join(''));
    };
    TS.prototype.writeContents = function () {
        var output = this.writeHeader();
        var doc = this.doc;
        var namespace = this.namespace;
        var prefix;
        var exportCount = 0;
        output.push('');
        output = output.concat(this.exportSourceList(namespace.sourceList));
        for (var _i = 0, _a = namespace.typeList.slice(0).sort(function (a, b) { return a.safeName.localeCompare(b.safeName); }); _i < _a.length; _i++) {
            var type = _a[_i];
            if (!type)
                continue;
            var isExported = (namespace.typeStateList[type.surrogateKey] == Namespace_1.TypeState.exported);
            output.push(this.writeType(type, isExported));
            if (isExported)
                ++exportCount;
        }
        for (var _b = 0, _c = doc.childList; _b < _c.length; _b++) {
            var child = _c[_b];
            var outElement = this.writeMember(child, true);
            if (outElement) {
                output.push(outElement);
                ++exportCount;
            }
        }
        if (!exportCount)
            output.push('export {};');
        output.push('');
        return (output.join('\n'));
    };
    TS.prototype.getOutName = function (name) {
        return (name + '.d.ts');
    };
    return TS;
}(Exporter_1.Exporter));
exports.TS = TS;
