// This file is part of cxsd, copyright (c) 2015-2016 BusFaster Ltd.
// Released under the MIT license, see LICENSE.
"use strict";
var types = require('./types');
var schema = require('../schema');
function mergeDuplicateTypes(typeList) {
    if (typeList.length < 2)
        return (typeList);
    var tbl = {};
    for (var _i = 0, typeList_1 = typeList; _i < typeList_1.length; _i++) {
        var type = typeList_1[_i];
        tbl[type.surrogateKey] = type;
    }
    return (Object.keys(tbl).map(function (key) { return tbl[key]; }));
}
/** Handle substitution groups. */
function expandSubstitutes(element) {
    // Filter out types of abstract elements.
    var elementList = element.isAbstract() ? [] : [element];
    if (element.substituteList) {
        elementList = elementList.concat.apply(elementList, element.substituteList.map(expandSubstitutes));
    }
    return (elementList);
}
/** Group elements by name and list all their different types. */
function mergeDuplicateElements(specList) {
    var groupTbl = {};
    for (var _i = 0, specList_1 = specList; _i < specList_1.length; _i++) {
        var spec = specList_1[_i];
        var element = spec.item;
        var group = groupTbl[element.getScope().namespace.id + ':' + element.name];
        if (!group) {
            group = {
                min: spec.min,
                max: spec.max,
                item: element,
                typeList: element.getTypes()
            };
            groupTbl[element.name] = group;
        }
        else {
            if (group.min > spec.min)
                group.min = spec.min;
            if (group.max < spec.max)
                group.max = spec.max;
            group.typeList = group.typeList.concat(element.getTypes());
        }
    }
    return (Object.keys(groupTbl).sort().map(function (name) { return groupTbl[name]; }));
}
function exportMember(group, parentScope, namespace) {
    var member = group.item;
    var scope = member.getScope();
    var otherNamespace = scope.namespace;
    var outMember = new schema.Member(member.name, group.min, group.max);
    outMember.comment = scope.getComments();
    if (otherNamespace != parentScope.namespace) {
        outMember.namespace = schema.Namespace.register(otherNamespace.id, otherNamespace.name, otherNamespace.short);
    }
    else
        outMember.namespace = namespace;
    outMember.typeList = mergeDuplicateTypes(group.typeList).map(function (type) {
        var outType = type.getOutType();
        var qName = type.qName;
        if (!qName && !type.name && !type.exported) {
            // Anonymous type defined only within this element.
            outType.containingMember = outMember;
            // Look through parent scopes for a containing type,
            // If the member was referenced from another namespace,
            // its scope points to definition in that namespace.
            var parentType = scope.getParentType(otherNamespace);
            if (parentType) {
                outType.containingType = parentType.getOutType();
            }
            exportType(type, outMember.namespace);
        }
        return (outType);
    });
    return (outMember);
}
function exportAttributes(scope, namespace) {
    var attributeTbl = scope.dumpAttributes();
    var attributeList = [];
    for (var _i = 0, _a = Object.keys(attributeTbl).sort(); _i < _a.length; _i++) {
        var key = _a[_i];
        var spec = attributeTbl[key];
        var attribute = spec.item;
        var group = {
            min: spec.min,
            max: spec.max,
            item: attribute,
            typeList: attribute.getTypes()
        };
        var outAttribute = exportMember(group, scope, namespace);
        if (outAttribute)
            attributeList.push(outAttribute);
    }
    return (attributeList);
}
function exportChildren(scope, namespace) {
    var elementTbl = scope.dumpElements();
    var childList = [];
    var specList = [];
    for (var _i = 0, _a = Object.keys(elementTbl); _i < _a.length; _i++) {
        var key = _a[_i];
        var spec = elementTbl[key];
        var min = spec.min;
        var max = spec.max;
        var substituteList = expandSubstitutes(spec.item);
        // If there are several alternatives, no specific one is mandatory.
        if (substituteList.length > 1)
            min = 0;
        for (var _b = 0, substituteList_1 = substituteList; _b < substituteList_1.length; _b++) {
            var element = substituteList_1[_b];
            specList.push({
                min: min,
                max: max,
                item: element
            });
        }
    }
    var groupList = mergeDuplicateElements(specList);
    for (var _c = 0, groupList_1 = groupList; _c < groupList_1.length; _c++) {
        var group = groupList_1[_c];
        var outChild = exportMember(group, scope, namespace);
        if (outChild)
            childList.push(outChild);
    }
    return (childList);
}
function exportType(type, namespace) {
    var scope = type.getScope();
    var comment = scope.getComments();
    type.exported = true;
    var outType = type.getOutType();
    outType.comment = comment;
    outType.bytePos = type.bytePos;
    var parentPrimitive = type.getParent(types.Primitive, true);
    if (parentPrimitive) {
        var primitiveList;
        var parentSimple = type.getParent(types.SimpleType, false);
        if (parentSimple) {
            // If parent is restricted to enumerated alternatives, output those instead.
            primitiveList = parentSimple.getEnumerationList();
            if (primitiveList)
                primitiveList = primitiveList.map(function (content) { return '"' + content + '"'; });
        }
        outType.primitiveList = primitiveList;
        outType.literalType = parentPrimitive.getOutType();
    }
    var parent = type.parent;
    if (parent instanceof types.TypeBase && parent != parentPrimitive) {
        outType.parent = parent.getOutType();
    }
    outType.attributeList = exportAttributes(scope, namespace);
    outType.childList = exportChildren(scope, namespace);
    return (outType);
}
/** Export parsed xsd into a simpler internal schema format. */
function exportNamespace(namespace) {
    var outNamespace = schema.Namespace.register(namespace.id, namespace.name, namespace.short);
    var doc = outNamespace.doc;
    if (!doc) {
        var scope = namespace.getScope();
        var sourceList = namespace.getSourceList();
        var importTbl = {};
        for (var _i = 0, sourceList_1 = sourceList; _i < sourceList_1.length; _i++) {
            var source = sourceList_1[_i];
            var namespaceRefTbl = source.getNamespaceRefs();
            for (var _a = 0, _b = Object.keys(namespaceRefTbl); _a < _b.length; _a++) {
                var name = _b[_a];
                var otherNamespace = namespaceRefTbl[name];
                outNamespace.addRef(name, schema.Namespace.register(otherNamespace.id, otherNamespace.name, otherNamespace.short));
                importTbl[otherNamespace.id] = otherNamespace;
            }
        }
        outNamespace.sourceList = sourceList.map(function (source) { return source.url; }).sort();
        var typeTbl = scope.dumpTypes();
        for (var _c = 0, _d = Object.keys(typeTbl); _c < _d.length; _c++) {
            var key = _d[_c];
            outNamespace.exportType(exportType(typeTbl[key].item, outNamespace));
        }
        doc = new schema.Type();
        doc.namespace = outNamespace;
        doc.attributeList = exportAttributes(scope, outNamespace);
        doc.childList = exportChildren(scope, outNamespace);
        outNamespace.doc = doc;
        for (var _e = 0, _f = Object.keys(importTbl); _e < _f.length; _e++) {
            var namespaceId = _f[_e];
            exportNamespace(importTbl[namespaceId]);
        }
    }
    return (doc);
}
exports.exportNamespace = exportNamespace;
