// This file is part of cxsd, copyright (c) 2015-2016 BusFaster Ltd.
// Released under the MIT license, see LICENSE.
"use strict";
/** Scope handles looking up references by type and name, and binding member
  * types and elements to types or namespaces. */
var Scope = (function () {
    function Scope(parent, namespace) {
        this.visible = {};
        this.expose = {};
        if (!namespace && parent)
            namespace = parent.namespace;
        this.parent = parent;
        this.namespace = namespace;
    }
    Scope.prototype.addString = function (name, type, target, min, max) {
        var visibleTbl = this.visible[type];
        if (!visibleTbl) {
            visibleTbl = {};
            this.visible[type] = visibleTbl;
        }
        visibleTbl[name] = target;
        var exposeTbl = this.expose[type];
        if (!exposeTbl) {
            exposeTbl = {};
            this.expose[type] = exposeTbl;
        }
        if (exposeTbl[name]) {
            // For sequences, sum occurrence counts among matching element types.
            min += exposeTbl[name].min;
            max += exposeTbl[name].max;
        }
        exposeTbl[name] = {
            min: min,
            max: max,
            item: target
        };
    };
    Scope.prototype.add = function (name, type, target, min, max) {
        this.addString(name.nameFull, type, target, min, max);
    };
    Scope.prototype.addToParent = function (name, type, target, min, max) {
        this.parent.add(name, type, target, min, max);
    };
    Scope.prototype.addAllToParent = function (type, min, max, target) {
        if (min === void 0) { min = 1; }
        if (max === void 0) { max = 1; }
        // Check if there's anything to add.
        if (!this.visible[type])
            return;
        if (!target)
            target = this;
        target = target.parent;
        var exposeTbl = this.expose[type];
        for (var _i = 0, _a = Object.keys(exposeTbl); _i < _a.length; _i++) {
            var name = _a[_i];
            var spec = exposeTbl[name];
            // TODO: If target is a choice, it must take the overall min and max.
            target.addString(name, type, spec.item, spec.min * min, spec.max * max);
        }
    };
    Scope.prototype.addComments = function (commentList) {
        this.commentList = (this.commentList || []).concat(commentList);
    };
    Scope.prototype.addCommentsToGrandParent = function (commentList) {
        this.parent.parent.addComments(commentList);
    };
    Scope.prototype.getComments = function () {
        if (!this.commentList)
            return (null);
        // Convert line breaks.
        return (this.commentList.join('').replace(/\r\n?|\n/g, '\n'));
    };
    Scope.prototype.lookup = function (name, type) {
        var scope = this;
        var nameFull = name.nameFull;
        var nameWild = '*:' + name.name;
        if (name.namespace && name.namespace != this.namespace) {
            scope = name.namespace.getScope();
        }
        var iter = 100;
        while (scope && --iter) {
            if (scope.visible[type]) {
                var result = scope.visible[type][nameFull] || scope.visible[type][nameWild];
                if (result)
                    return (result);
            }
            scope = scope.parent;
        }
        console.log('Missing ' + type + ': ' + name.name);
        return (null);
    };
    // Types
    Scope.prototype.setType = function (type) {
        // TODO: set to some invalid value if called more than once.
        if (!this.type && this.namespace.getScope() != this)
            this.type = type;
    };
    Scope.prototype.setParentType = function (type) {
        this.parent.setType(type);
    };
    Scope.prototype.getParentType = function (namespace) {
        for (var parent = this.parent; parent && parent.namespace == namespace; parent = parent.parent) {
            if (parent.type)
                return (parent.type);
        }
        return (null);
    };
    Scope.prototype.getType = function () { return (this.type); };
    Scope.prototype.dumpTypes = function () {
        return (this.expose['type'] || {});
    };
    Scope.prototype.dumpElements = function () {
        return (this.expose['element'] || {});
    };
    Scope.prototype.dumpAttributes = function () {
        return (this.expose['attribute'] || {});
    };
    Scope.prototype.hasAttributes = function () {
        var tbl = this.expose['attribute'];
        for (var key in tbl)
            if (tbl.hasOwnProperty(key))
                return (true);
        return (false);
    };
    return Scope;
}());
exports.Scope = Scope;
